-- BattleFlow.lua

local L = LibStub("AceLocale-3.0"):GetLocale("AshranBuddy")

local AshranBuddy = _G.AshranBuddy
if (not AshranBuddy) then
	DEFAULT_CHAT_FRAME:AddMessage(format("Module '%s' requires AshranBuddy", "Battle Flow"))
end

-- Registration
local BattleFlow = AshranBuddy:RegisterModule("BattleFlow")
LibStub("AceTimer-3.0"):Embed(BattleFlow)

-- Locals
local Points = { L["Grand Marshal Tremblade"], L["Archmage Overwatch"], L["Tremblade's Vanguard"], L["The Crossroads"], L["Volrath's Advance"], L["Emberfall Tower"], L["High Warlord Volrath"] }
local States = { Intermission = -1, Contested = 0, Alliance = 1, Horde = 2 }
local Colors = { }
Colors[States.Alliance] = { Red = (20 / 255), Green = (69 / 255), Blue = (133 / 255) }
Colors[States.Contested] = { Red = (180 / 210), Green = (180 / 255), Blue = (0 / 255) }
Colors[States.Horde] = { Red = (191 / 255), Green = (27 / 255), Blue = (27 / 255) }
Colors[States.Intermission] = { Red = 0.5, Green = 0.5, Blue = 0.5 }

local Previous = { States.Intermission, States.Intermission, States.Intermission, States.Intermission, States.Intermission, States.Intermission, States.Intermission }
local CanSendUpdate = true
local CanReceiveUpdate = false
local TooltipsCreated = false

function BattleFlow:OnInitialize()
	-- Battle Flow operates using localised text returned by GetWorldStateUIInfo(i)
	-- The text used must match exactly otherwise the majority of this module will not function
	-- Each language will be enabled manually until all Locales are available
	if (not self:IsCompatible()) then
		CanSendUpdate = false
		return
	end

	self:InitializeComponent(25)
	
	if (not AshranBuddy:IsInAshran()) then
		self.component:Hide()
	end
end

function BattleFlow:OnEnable()
	-- Battle Flow operates using localised text returned by GetWorldStateUIInfo(i)
	-- The text used must match exactly otherwise the majority of this module will not function
	-- Each language will be enabled manually until all Locales are available
	if (not self:IsCompatible()) then
		CanSendUpdate = false
		return
	end
	
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "OnPlayerEnteringWorld")
	self:RegisterEvent("UPDATE_WORLD_STATES", "OnUpdateWorldStates")
	self:RegisterEvent("WORLD_STATE_UI_TIMER_UPDATE", "OnUpdateWorldStates")
	self:RegisterEvent("ZONE_CHANGED", "OnZoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnZoneChangedNewArea")
	
	if (not AshranBuddy.db.profile.BattleFlow.locked) then
		BattleFlow:Unlock()
	end
	
	if (AshranBuddy:IsInAshran()) then
		self.component:Show()
		WorldStateAlwaysUpFrame:Hide()
		CanSendUpdate = true
	else
		if (not AshranBuddy.db.profile.BattleFlow.locked) then
			BattleFlow:Unlock()
			self.component:Show()
		else
			self.component:Hide()
		end
		WorldStateAlwaysUpFrame:Show()
		CanSendUpdate = false
	end
end

function BattleFlow:OnDisable()
	self:UnregisterAllEvents()
	self.component:Hide()
	CanSendUpdate = false
	
	WorldStateAlwaysUpFrame:Show()
end

-- Battle Flow will only operate correctly for English and French game clients at this stage.
function BattleFlow:IsCompatible()
	local locale = GetLocale()
	return locale == "enUS" or locale == "enGB" or locale == "frFR"
end

function BattleFlow:OnChatCommand(input)
	if (input == "unlock") then
		BattleFlow:Unlock()
		AshranBuddy.db.profile.BattleFlow.locked = false
	elseif (input == "lock") then
		BattleFlow:Lock()
		AshranBuddy.db.profile.BattleFlow.locked = true
	end
end

function BattleFlow:OnUpdateWorldStates(event)
	if (AshranBuddy:IsInAshran()) then
		local states, slain, timer, event, capture = self:GetWorldStates()
		self:SetWorldStates(states, slain, timer, event, capture)
		self:SetTooltips()
		self:SuppressDefaultCaptureBar()
	end
end

function BattleFlow:OnZoneChangedNewArea(event)
	if (AshranBuddy:IsInAshran()) then
		self.component:Show()
		WorldStateAlwaysUpFrame:Hide()		
	else
		if (AshranBuddy.db.profile.BattleFlow.locked == false) then
			self:Unlock()
		else
			self.component:Hide()
		end
		WorldStateAlwaysUpFrame:Show()	
	end
end

function BattleFlow:OnZoneChanged(event)
	-- Check if in middle lane, if not allow others to
	-- sync to us the main lane progress
end

function BattleFlow:OnPlayerEnteringWorld(event)
	if (AshranBuddy:IsInAshran()) then
		self.component:Show()
		WorldStateAlwaysUpFrame:Hide()	
		self:OnUpdateWorldStates()
	else
		if (AshranBuddy.db.profile.EventTimer.locked == false) then
			self:Unlock()
		else
			self.component:Hide()
		end
		WorldStateAlwaysUpFrame:Show()	
	end
end

function BattleFlow:Lock()
	if (not AshranBuddy:IsInAshran()) then
		self.component:Hide()
	end
	self.handle:Hide()
	
	local point, relativeTo, relativePoint, x, y = self.component:GetPoint(1)
	AshranBuddy.db.profile.BattleFlow.point = point
	AshranBuddy.db.profile.BattleFlow.relativePoint = relativePoint
	AshranBuddy.db.profile.BattleFlow.x = x
	AshranBuddy.db.profile.BattleFlow.y = y
end

function BattleFlow:Unlock()
	self.component:Show()
	self.handle:Show()
end

function BattleFlow:InitializeComponent(size)
	local count = #Points
	local offset = 5
	local width = (size * count) + (offset * (count - 1))
	
	self.controlPoints = { }
	local profile = AshranBuddy.db.profile.BattleFlow
	
	local frame = self.component
	frame:SetParent(UIParent)
	frame:SetSize(width, size)
	frame:SetPoint(profile.point, UIParent, profile.relativePoint, profile.x, profile.y)
	frame:SetToplevel(true)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	
	local x = 0
	for i = 1, #Points do
		self.controlPoints[i] = self:CreateControlPointFrame("Point"..i, Points[i], frame, size, size, x, 0)
		x = x + size + offset
	end
	
	self:CreateFactionMarkerFrame(frame)
	self:CreateSlainCounterFrame(frame)
	self:CreateTimerFrame(frame)
	self:CreateEventBarFrame(frame, width)
	self:CreateCaptureBarFrame(frame, width)
	self:CreateBossTooltipFrame(frame)
	
	frame:SetScript("OnHide", function(self)
		for i = 1, #BattleFlow.controlPoints do
			BattleFlow.controlPoints[i].state = States.Intermission
		end
		self.SlainCounter.Alliance:SetText("")
		self.SlainCounter.Horde:SetText("")
		self.Timer:SetText("")
	end)
	
	-- Handle
	local handle = CreateFrame("Frame", nil, frame)
	handle:SetSize(width + 200, size + 45)
	handle:SetPoint("TOPLEFT", frame, "TOPLEFT", -100, 0)
	handle:Hide()

	local handleBox = handle:CreateTexture(nil, "OVERLAY", nil, -1)
	handleBox:SetTexture(0, 0, 0, 0.6)
	handleBox:SetSize(width + 200, size + 45)
	handleBox:SetPoint("TOPLEFT")
	
	local handleFontString = handle:CreateFontString("HandleLabel", "OVERLAY", "TextStatusBarText")
	handleFontString:SetPoint("CENTER")
	handleFontString:SetText(L["Ashran Buddy\n\nYou MUST Type /ab lock or\n/ashranbuddy lock to activate.\nPoints won't show until your are in Ashran."])

	handle:SetScript("OnMouseDown", function(self)
		print(AshranBuddy.db.profile.BattleFlow.locked)
		if (AshranBuddy.db.profile.BattleFlow.locked == false) then
			self:GetParent():StartMoving()
		end
	end)
	handle:SetScript("OnMouseUp", function(self)
		if (AshranBuddy.db.profile.BattleFlow.locked == false) then
			self:GetParent():StopMovingOrSizing()
		end
	end)
	
	self.handle = handle
end

function BattleFlow:CreateControlPointFrame(name, title, parent, width, height, x, y)
	local color = Colors[States.Intermission]

	local controlPoint = CreateFrame("Frame", name, parent)
	controlPoint:SetSize(width, height)
	controlPoint:SetPoint("TOPLEFT", parent, "TOPLEFT", x, y)

	controlPoint.Texture = controlPoint:CreateTexture(nil, "BACKGROUND")
	controlPoint.Texture:SetTexture(color.Red, color.Green, color.Blue, 1)
	controlPoint.Texture:SetSize(width, height)
	controlPoint.Texture:SetPoint("TOPLEFT")
	
	controlPoint.title = title
	controlPoint.tooltip = nil
	controlPoint.state = States.Contested

	controlPoint.GetTitle = function(self)
		return self.title
	end
	controlPoint.GetTooltip = function(self)
		return self.tooltip
	end
	controlPoint.SetTooltip = function(self, value)
		self.tooltip = value
	end
	controlPoint.GetState = function(self)
		return self.state
	end
	controlPoint.SetState = function(self, state)
		local color = Colors[state]
		self.Texture:SetTexture(color.Red, color.Green, color.Blue, 1)
		self.state = state
	end
	controlPoint:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self)
		GameTooltip:SetText(self.title)
		if (self.tooltip ~= nil) then
			GameTooltip:AddLine(self.tooltip, 1, 1, 1, true)
		end
        GameTooltip:Show()
	end)
	controlPoint:SetScript("OnLeave", function(self)
		GameTooltip_Hide()
	end)

	return controlPoint
end

function BattleFlow:CreateEventBarFrame(parent, width)
	local height = 5
	local eventBar = CreateFrame("Frame", nil, parent)
	eventBar:SetSize(width, height)
	eventBar:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", 0, -(height * 2))

	self:CreateEventBarFactionFrame(eventBar, "Alliance", width, height, Colors[States.Alliance], "RIGHT", "LEFT", -20)
	
	local extent = 4
	eventBar.Divider = CreateFrame("Frame", nil, eventBar)
	eventBar.Divider:SetSize(extent, height + (extent * 2))
	eventBar.Divider:SetPoint("CENTER", eventBar, "CENTER", 0, 0)
	eventBar.Divider:SetFrameStrata("HIGH")
	eventBar.Divider.Texture = eventBar.Divider:CreateTexture(nil, "BACKGROUND")
	eventBar.Divider.Texture:SetTexture(1, 1, 1, 1)
	eventBar.Divider.Texture:SetSize(extent, height + (extent * 2))
	eventBar.Divider.Texture:SetPoint("CENTER")
	
	self:CreateEventBarFactionFrame(eventBar, "Horde", width, height, Colors[States.Horde], "LEFT", "RIGHT", 20)
	
	eventBar:SetScript("OnShow", function(self)
		local parent = self:GetParent()
		parent.Timer:ClearAllPoints()
		parent.Timer:SetPoint("TOP", parent, "BOTTOM", 0, -15)
	end)
	eventBar:SetScript("OnHide", function(self)
		local parent = self:GetParent()
		parent.Timer:ClearAllPoints()
		parent.Timer:SetPoint("TOP", parent, "BOTTOM")
	end)
	
	eventBar:Hide()
	
	parent.EventBar = eventBar
end

function BattleFlow:CreateEventBarFactionFrame(eventBar, faction, width, height, color, anchor1, anchor2, offset)
	eventBar[faction] = eventBar:CreateTexture(nil, "BACKGROUND")
	eventBar[faction]:SetTexture(color.Red, color.Green, color.Blue, 1)
	eventBar[faction]:SetSize(width / 2, height)
	eventBar[faction]:SetPoint(anchor1, eventBar, "CENTER")
	
	eventBar[faction].Text = eventBar:CreateFontString(nil , "BACKGROUND", "GameFontNormalSmall")
	eventBar[faction].Text:SetSize(200, 25)
	eventBar[faction].Text:SetJustifyH(anchor1)
	eventBar[faction].Text:SetPoint(anchor1, eventBar, anchor2, offset, 0)
	eventBar[faction].Text:SetText("")
	
	eventBar[faction].SetMinMaxValues = function(self, minValue, maxValue)
		self.minValue = minValue
		self.maxValue = maxValue
	end
	eventBar[faction].SetValue = function(self, value)
		self.value = value
		if (value == 0) then
			self:Hide()
		else
			self:Show()
		end
		if (self.maxValue > 0) then
			self:SetWidth((self.value / self.maxValue) * self.maxWidth)
		end
	end
	
	eventBar[faction].minValue = 0
	eventBar[faction].maxValue = 0
	eventBar[faction].value = 0
	eventBar[faction].maxWidth = width / 2
end

function BattleFlow:CreateCaptureBarFrame(parent, width)
	local color = Colors[States.Alliance]
	local padding = 10
	local height = 5
	local extent = 4

	local captureBar = CreateFrame("Frame", nil, parent)
	captureBar:SetSize(width, height)
	captureBar:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", 0, -(height * 2))
	
	captureBar.maxWidth = width
	
	captureBar.Alliance = captureBar:CreateTexture(nil, "BACKGROUND")
	captureBar.Alliance:SetTexture(color.Red, color.Green, color.Blue, 1)
	captureBar.Alliance:SetSize(padding, height)
	captureBar.Alliance:SetPoint("LEFT", captureBar, "LEFT")
	
	color = Colors[States.Horde]
	
	captureBar.Horde = captureBar:CreateTexture(nil, "BACKGROUND")
	captureBar.Horde:SetTexture(color.Red, color.Green, color.Blue, 1)
	captureBar.Horde:SetSize(padding, height)
	captureBar.Horde:SetPoint("RIGHT", captureBar, "RIGHT")
	
	color = Colors[States.Intermission]
	
	captureBar.Neutral = captureBar:CreateTexture(nil, "BACKGROUND")
	captureBar.Neutral:SetTexture(color.Red, color.Green, color.Blue, 1)
	captureBar.Neutral:SetSize(width - (padding * 2), height)
	captureBar.Neutral:SetPoint("CENTER", captureBar, "CENTER")
	
	captureBar.Divider = CreateFrame("Frame", nil, captureBar)
	captureBar.Divider:SetSize(extent, height + (extent * 2))
	captureBar.Divider:SetPoint("LEFT", captureBar, "LEFT", (width / 2) - (extent / 2), 0)
	captureBar.Divider:SetFrameStrata("HIGH")
	captureBar.Divider.Texture = captureBar.Divider:CreateTexture(nil, "BACKGROUND")
	captureBar.Divider.Texture:SetTexture(1, 1, 1, 1)
	captureBar.Divider.Texture:SetSize(extent, height + (extent * 2))
	captureBar.Divider.Texture:SetPoint("CENTER")
	captureBar.SetValue = function(self, value)
		self.value = value
		self.Divider:ClearAllPoints()
		self.Divider:SetPoint("LEFT", self, "LEFT", (((100 - self.value) / 100) * self.maxWidth) - 2, 0)
	end
	
	captureBar:SetScript("OnShow", function(self)
		local parent = self:GetParent()
		if (parent.EventBar:IsShown()) then
			self:ClearAllPoints()
			self:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", 0, -25)
			parent.Timer:ClearAllPoints()
			parent.Timer:SetPoint("TOP", parent, "BOTTOM", 0, -30)
		else
			parent.Timer:ClearAllPoints()
			parent.Timer:SetPoint("TOP", parent, "BOTTOM")
		end
	end)
	captureBar:SetScript("OnHide", function(self)
		local parent = self:GetParent()
		self:ClearAllPoints()
		self:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", 0, -10)
		if (parent.EventBar:IsShown()) then
			parent.Timer:ClearAllPoints()
			parent.Timer:SetPoint("TOP", parent, "BOTTOM", 0, -15)
		else
			parent.Timer:ClearAllPoints()
			parent.Timer:SetPoint("TOP", parent, "BOTTOM")
		end
	end)
	
	captureBar:Hide()
	
	parent.CaptureBar = captureBar
end

function BattleFlow:CreateFactionMarkerFrame(parent)
	local alliance = CreateFrame("Frame", nil, parent)
	alliance:SetSize(25, 25)
	alliance:SetPoint("RIGHT", self.controlPoints[1], "LEFT", 12, 0)
	alliance:SetFrameStrata("HIGH")
	
	alliance.Texture = alliance:CreateTexture(nil, "BACKGROUND")
	alliance.Texture:SetTexture("Interface\\TargetingFrame\\UI-PVP-Alliance")
	alliance.Texture:SetSize(40, 40)
	alliance.Texture:SetPoint("LEFT", -1, -7)
	
	local horde = CreateFrame("Frame", nil, parent)
	horde:SetSize(25, 25)
	horde:SetPoint("LEFT", self.controlPoints[7], "RIGHT", -12, 0)
	horde:SetFrameStrata("HIGH")
	
	horde.Texture = horde:CreateTexture(nil, "BACKGROUND")
	horde.Texture:SetTexture("Interface\\TargetingFrame\\UI-PVP-Horde")
	horde.Texture:SetSize(40, 40)
	horde.Texture:SetPoint("LEFT", -1, -8)
end

function BattleFlow:CreateSlainCounterFrame(parent)
	parent.SlainCounter = { }

	local allianceFrame = CreateFrame("Frame", nil, parent)
	allianceFrame:SetSize(100, 25)
	allianceFrame:SetPoint("RIGHT", self.controlPoints[1], "LEFT", -20, 0)
	
	allianceFrame.Counter = allianceFrame:CreateFontString(nil , "BACKGROUND", "GameFontNormalSmall")
	allianceFrame.Counter:SetSize(100, 25)
	allianceFrame.Counter:SetJustifyH("RIGHT")
	allianceFrame.Counter:SetPoint("RIGHT", self.controlPoints[1], "LEFT", -20, 0)
	allianceFrame.Counter:SetText("")
	allianceFrame.Counter.Tooltip = ""
	
	allianceFrame:SetScript("OnEnter", function(self)
		if (self.Counter.ToolTip) then
			GameTooltip:SetOwner(self)
			GameTooltip:SetText(self.Counter.Tooltip)
		end
	end)
	allianceFrame:SetScript("OnLeave", function(self)
		GameTooltip_Hide()
	end)
	
	parent.SlainCounter.Alliance = allianceFrame.Counter
	
	local hordeFrame = CreateFrame("Frame", nil, parent)
	hordeFrame:SetSize(100, 25)
	hordeFrame:SetPoint("LEFT", self.controlPoints[7], "RIGHT", 20, 0)
	
	hordeFrame.Counter = hordeFrame:CreateFontString(nil , "BACKGROUND", "GameFontNormalSmall")
	hordeFrame.Counter:SetSize(100, 25)
	hordeFrame.Counter:SetJustifyH("LEFT")
	hordeFrame.Counter:SetPoint("LEFT", self.controlPoints[7], "RIGHT", 20, 0)
	hordeFrame.Counter:SetText("")
	hordeFrame.Counter.Tooltip = ""
	
	hordeFrame:SetScript("OnEnter", function(self)
		if (self.Counter.ToolTip) then
			GameTooltip:SetOwner(self)
			GameTooltip:SetText(self.Counter.Tooltip)
		end
	end)
	hordeFrame:SetScript("OnLeave", function(self)
		GameTooltip_Hide()
	end)
	
	parent.SlainCounter.Horde = hordeFrame.Counter
end

function BattleFlow:CreateTimerFrame(parent)
	local timerFrame = CreateFrame("Frame", nil, parent)
	timerFrame:SetSize(200, 25)
	timerFrame:SetPoint("TOP", parent, "BOTTOM")

	timerFrame.Timer = timerFrame:CreateFontString(nil , "BACKGROUND", "GameFontNormalSmall")
	timerFrame.Timer:SetSize(200, 25)
	timerFrame.Timer:SetPoint("TOP", parent, "BOTTOM")
	timerFrame.Timer:SetText("")
	timerFrame.Timer.Tooltip = ""
	
	timerFrame:SetScript("OnEnter", function(self)
		if (self.Timer.ToolTip) then
			GameTooltip:SetOwner(self)
			GameTooltip:SetText(self.Timer.Tooltip)
		end
	end)
	timerFrame:SetScript("OnLeave", function(self)
		GameTooltip_Hide()
	end)
	
	parent.Timer = timerFrame.Timer
end

function BattleFlow:CreateBossTooltipFrame(parent)
	local bossTooltip = CreateFrame("Frame", nil, parent)
	bossTooltip:SetSize(200, 25)
	bossTooltip:SetPoint("TOP", parent, "BOTTOM")

	bossTooltip.Text = bossTooltip:CreateFontString(nil , "BACKGROUND", "GameFontNormalSmall")
	bossTooltip.Text:SetSize(200, 25)
	bossTooltip.Text:SetPoint("TOP", parent, "BOTTOM")
	bossTooltip.Text:SetText("")
	
	bossTooltip:SetScript("OnShow", function(self)
		local parent = self:GetParent()
		parent.Timer:ClearAllPoints()
		parent.Timer:SetPoint("TOP", parent, "BOTTOM", 0, -15)
	end)
	bossTooltip:SetScript("OnHide", function(self)
		local parent = self:GetParent()
		if (not parent.EventBar:IsShown() and not parent.CaptureBar:IsShown()) then
			parent.Timer:ClearAllPoints()
			parent.Timer:SetPoint("TOP", parent, "BOTTOM")
		end
	end)
	
	bossTooltip:Hide()
	
	parent.BossTooltip = bossTooltip
end

function BattleFlow:GetWorldStates()
	local states = { States.Intermission, States.Intermission, States.Intermission, States.Intermission, States.Intermission, States.Intermission, States.Intermission }
	local slain = nil
	local timer = nil
	local event = nil
	local capture = nil
	
	local count = GetNumWorldStateUI()
	for i = 1, count do
		local uiType, state, hidden, text, icon, dynamicIcon, tooltip, dynamicTooltip, extendedUI, extendedUIState1, extendedUIState2, extendedUIState3 = GetWorldStateUIInfo(i)
		if (uiType == 3 and state > 0) then
			if (timer == nil) then
				timer = { }
			end
			timer.Text = text
			timer.Tooltip = tooltip
		elseif (uiType == 0 and state > 0) then
			local frame = self.component
			local index = table.indexof(Points, text)
			if (index ~= -1) then
				states[index] = States.Contested
			elseif (text:find(L["Enemies Slain"])) then
				if (slain == nil) then
					slain = { }
				end
				local i = #slain + 1
				slain[i] = { Faction = "", Text = "" }
				
				local startIndex, endIndex = text:find(": ")
				slain[i].Faction = "Horde"
				if (icon:find("Alliance")) then
					slain[i].Faction = "Alliance"
				end
				slain[i].Text = text:sub(endIndex + 1)
				slain[i].Tooltip = tooltip
			elseif (text:find(L["Ore Collected"]) or text:find(L["Risen Spirits Captured"]) or text:find(L["Laps"]) or text:find(L["Score"])) then
				if (event == nil) then
					event = { }
				end
				local i = #event + 1
				event[i] = { Faction = "", Value = 0, MaxValue = 0, Text = "" }

				event[i].Faction = "Horde"
				if (icon:find("Alliance")) then
					event[i].Faction = "Alliance"
				end
				
				local startIndex = string.lastindexof(text, ": ")
				local endIndex = string.lastindexof(text, "/")
				local value = text:sub(startIndex + 1, endIndex - 1)
				local maxValue = text:sub(endIndex + 1)
				
				event[i].MaxValue = tonumber(maxValue)
				event[i].Value = tonumber(value)
				event[i].Text = text
			elseif (extendedUI == "CAPTUREPOINT") then
				capture = tonumber(extendedUIState1)
			else
				if (timer == nil) then
					timer = { }
				end
				timer.Text = text
				timer.Tooltip = tooltip
			end
		end
		if (state == 0 and (text:find(L["Ore Collected"]) or text:find(L["Risen Spirits Captured"]) or text:find(L["Laps"]) or text:find(L["Score"]))) then
			event = nil
		elseif (state == 0 and extendedUI == "CAPTUREPOINT" and extendedUIState2 > 0) then
			capture = nil
		end
	end
	if (table.contains(states, States.Contested)) then
		local faction = States.Alliance
		for i = 1, #states do
			if (states[i] == States.Contested) then
				faction = States.Horde
			else
				states[i] = faction
			end
		end
	end
	return states, slain, timer, event, capture
end

function BattleFlow:SetWorldStates(states, slain, timer, event, capture)
	local frame = self.component
	-- States
	if (table.contains(states, States.Intermission)) then
		states = Previous
		CanReceiveUpdate = true
	elseif (#self.controlPoints == #states) then
		CanReceiveUpdate = false
		for i = 1, #self.controlPoints do
			local changed = false
			if (self.controlPoints[i]:GetState() ~= states[i]) then
				changed = true
			end
			self.controlPoints[i]:SetState(states[i])
			if (changed) then
				-- TODO: Animate transition, not a priority
			end
		end
		Previous = states
	else
		-- We may have left the main road
		-- let OnZoneChanged handle this
	end
	-- Slain
	if (slain ~= nil) then
		for i = 1, #slain do
			frame.SlainCounter[slain[i].Faction]:SetText(slain[i].Text)
			frame.SlainCounter[slain[i].Faction].Tooltip = slain[i].Tooltip
		end
	end
	-- Timer
	if (timer ~= nil) then
		frame.Timer:SetText(timer.Text)
		frame.Timer.Tooltip = timer.Tooltip
	end
	-- Boss Tooltip
	local point = table.indexof(states, States.Contested)
	if (point ~= -1 and (point == 1 or point == #Points)) then
		if (not frame.BossTooltip:IsShown() and not frame.EventBar:IsShown() and not frame.CaptureBar:IsShown()) then
			frame.BossTooltip:Show()
		end
		frame.BossTooltip.Text:SetText(self.controlPoints[point]:GetTooltip())
	else
		frame.BossTooltip:Hide()
	end
	-- Event
	if (event ~= nil) then
		if (not frame.EventBar:IsShown()) then
			frame.BossTooltip:Hide()
			frame.EventBar:Show()
		end
		for i = 1, #event do
			frame.EventBar[event[i].Faction]:SetMinMaxValues(0, event[i].MaxValue)
			frame.EventBar[event[i].Faction]:SetValue(event[i].Value)
			frame.EventBar[event[i].Faction].Text:SetText(event[i].Text)
		end
	elseif (event == nil and frame.EventBar:IsShown()) then
		frame.EventBar:Hide()
	end
	-- Capture
	if (capture ~= nil) then
		if (not frame.CaptureBar:IsShown()) then
			frame.BossTooltip:Hide()
			frame.CaptureBar:Show()
		end
		frame.CaptureBar:SetValue(capture)
	elseif (capture == nil and frame.CaptureBar:IsShown()) then
		frame.CaptureBar:Hide()
	end
	self:SendUpdate(states, slain, timer)
end

function BattleFlow:SetTooltips()
	if (not TooltipsCreated) then
		local count = GetNumWorldStateUI()
		for i = 1, count do
			local uiType, state, hidden, text, icon, dynamicIcon, tooltip, dynamicTooltip, extendedUI, extendedUIState1, extendedUIState2, extendedUIState3 = GetWorldStateUIInfo(i)
			local index = table.indexof(Points, text)
			if (index ~= -1) then
				self.controlPoints[index]:SetTooltip(tooltip)
				TooltipsCreated = true
			end
		end
	end	
end

function BattleFlow:OnMessageReceived(packet, distribution, sender)
	if (CanReceiveUpdate) then
		-- Only accept sync updates if the game client does not provide them
		if (packet.type == "update" and distribution == "RAID") then
			BattleFlow:ReceiveUpdate(sender, packet.data)
		end
	end
end

function BattleFlow:SendUpdate(states, slain, timer)
	if (IsInRaid() and CanSendUpdate and not CanReceiveUpdate) then
		-- An update can only be sent once every few (randomized) seconds
		-- This should cut down addon message spam
		local packet = { type = "update", data = { ["states"] = states, ["slain"] = slain, ["timer"] = timer } }
		self:SendMessage(packet, "RAID")
		
		CanSendUpdate = false
		self:ScheduleTimer("OnTimerCallback", math.random(0, 1))
	end
end

function BattleFlow:ReceiveUpdate(sender, data)
	if (data ~= nil) then
		local _, _, _, event, capture = self:GetWorldStates()
		self:SetWorldStates(data.states, data.slain, data.timer, event, capture)
	end
end

function BattleFlow:OnCreateOptions(options)
	if (self:IsCompatible()) then
		options.BattleFlow = {
			name = L["Battle Flow"],
			type = "group",
			order = 30,
			args = {
				header = {
					name = L["Battle Flow"],
					type = "header",
					order = 0
				},
				headerDescription = {
					name = L["Battle Flow description"],
					type = "description",
					order = 1
				},
				enabled = {
					name = L["Enabled"],
					desc = L["Enables the Battle Flow module."],
					type = "toggle",
					set = function(info, val) AshranBuddy.db.profile.BattleFlow.enabled = val; if (val) then self:Enable() else self:Disable() end end,
					get = function(info) return AshranBuddy.db.profile.BattleFlow.enabled end,
					order = 2
				},
				enabledDescription = {
					name = L["Enables the Battle Flow module."],
					type = "description",
					order = 3
				},
				locked = {
					name = L["Lock Battle Flow"],
					desc = L["Locks and unlocks the Battle Flow frame."],
					type = "toggle",
					set = function(info, val) AshranBuddy.db.profile.BattleFlow.locked = val; if (AshranBuddy.db.profile.BattleFlow.locked) then self:Lock() else self:Unlock() end end,
					get = function(info) return AshranBuddy.db.profile.BattleFlow.locked end,
					order = 4
				},
				lockedDescription = {
					name = L["Locks and unlocks the Battle Flow frame."],
					type = "description",
					order = 5
				},
				showTimer = {
					name = L["Reset Battle Flow Position"],
					type = "execute",
					func = function(info, val)
						local profile = AshranBuddy.db.profile
						profile.BattleFlow.point = "TOP"
						profile.BattleFlow.relativePoint = "TOP"
						profile.BattleFlow.x = 0
						profile.BattleFlow.y = -10
						self.component:ClearAllPoints()
						self.component:SetPoint(profile.BattleFlow.point, UIParent, profile.BattleFlow.relativePoint, profile.BattleFlow.x, profile.BattleFlow.y)
					end,
					order = 6
				},
			},
		}
	else
		options.BattleFlow = {
			name = L["Battle Flow"],
			type = "group",
			order = 30,
			args = {
				header = {
					name = L["Battle Flow"],
					type = "header",
					order = 0
				},
				headerDescription = {
					name = L["Battle Flow description"],
					type = "description",
					order = 1
				},
			},
		}
	end
end

function BattleFlow:OnTimerCallback()
	CanSendUpdate = true
end

function BattleFlow:SuppressDefaultCaptureBar()
	if (NUM_EXTENDED_UI_FRAMES) then
 		local captureBar
 		for i = 1, NUM_EXTENDED_UI_FRAMES do
 			captureBar = _G["WorldStateCaptureBar"..i]
 			if (captureBar) then
 				captureBar:Hide()
 			end
 		end
 	end
end